/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2015 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.annotation.editingsupport;

import java.util.Set;
import java.util.TreeSet;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EEnum;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.jface.viewers.ColumnViewer;
import org.eclipse.jface.viewers.EditingSupport;
import org.fortiss.tooling.base.annotation.AnnotationEntry;
import org.fortiss.tooling.base.annotation.valueprovider.DerivedAnnotationValueProviderBase;
import org.fortiss.tooling.base.annotation.valueprovider.DynamicInstanceAnnotationValueProviderBase;
import org.fortiss.tooling.base.annotation.valueprovider.EStructuralFeatureValueProviderBase;
import org.fortiss.tooling.base.annotation.valueprovider.IAnnotationValueProvider;
import org.fortiss.tooling.base.annotation.valueprovider.ValueProviderBase;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IDerivedAnnotation;
import org.fortiss.tooling.base.ui.annotation.view.generic.GenericAnnotationView;
import org.fortiss.tooling.kernel.utils.EcoreUtils;

/**
 * Factory to build {@link AnnotationEditingSupportBase}-based {@link EditingSupport} for the
 * {@link GenericAnnotationView}.
 * 
 * @author barner
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating RED Hash:
 */
public class EditingSupportFactory {

	/**
	 * Creates the {@link EditingSupport} that is directly based on
	 * {@link DerivedAnnotationValueProviderBase}.
	 */
	private static
			<T extends IDerivedAnnotation<?>, V extends DerivedAnnotationValueProviderBase<T>>
			EditingSupport createEditingSupport4DerivedAnnotationValueProviderBase(V valueProvider,
					ColumnViewer viewer, Class<? extends IAnnotatedSpecification> clazz,
					T specification, String instanceKey) throws Exception {

		if(valueProvider.isDerivedKey(instanceKey) && !valueProvider.isEditableByUser()) {
			return null;
		}

		return createEditingSupport4EStructuralFeatureValueProviderBase(
				(EStructuralFeatureValueProviderBase<T>)valueProvider, viewer, clazz,
				specification, instanceKey);
	}

	/**
	 * Creates the {@link EditingSupport} that is directly based on
	 * {@link EStructuralFeatureValueProviderBase}.
	 * 
	 * @param valueProvider
	 * @param specification
	 */
	private static
			<T extends IAnnotatedSpecification, V extends DynamicInstanceAnnotationValueProviderBase<T>>
			EditingSupport createEditingSupport4DynamicInstanceAnnotationValueProviderBase(
					V valueProvider, ColumnViewer viewer,
					Class<? extends IAnnotatedSpecification> clazz, T specification,
					String instanceKey) throws Exception {

		return new TextEditingSupport(viewer, clazz, instanceKey);
	}

	/**
	 * Creates the {@link EditingSupport} that is directly based on
	 * {@link EStructuralFeatureValueProviderBase}.
	 */
	private static
			<T extends IAnnotatedSpecification, V extends EStructuralFeatureValueProviderBase<T>>
			EditingSupport createEditingSupport4EStructuralFeatureValueProviderBase(
					V valueProvider, ColumnViewer viewer,
					Class<? extends IAnnotatedSpecification> clazz, T specification,
					String instanceKey) throws Exception {

		EStructuralFeature structuralFeature =
				valueProvider.getEStructuralFeature(specification, instanceKey);

		EClassifier eType = null;
		try {
			eType = structuralFeature.getEType();
		} catch(Exception e) {
			// Ignore
		}

		if(structuralFeature.getUpperBound() == 0) {
			throw new Exception(
					"EStructuralValueProvider: feature multiplicity == 0 is not supported.");
		}

		if(structuralFeature.getUpperBound() == 1) {

			if(((!(eType instanceof EEnum)) && (!(eType instanceof EClass))) ||
					((eType instanceof EClass) && (valueProvider.getEStructuralFeatureDescriptor(
							instanceKey).getEReferenceScope() == null))) {

				return createEditingSupport4ValueProviderBase((ValueProviderBase<T>)valueProvider,
						viewer, clazz, specification, instanceKey);
			}
			return new ComboBoxEditingSupport(viewer, clazz, specification, instanceKey,
					valueProvider.getEStructuralFeatureDescriptor(instanceKey));
		}

		if(eType instanceof EClass || eType instanceof EEnum) {
			return new MultiSelectionEditingSupport(viewer, clazz, specification, instanceKey,
					valueProvider.getEStructuralFeatureDescriptor(instanceKey));
		}

		throw new Exception(
				"For feature multiplicity > 1, EStructuralValueProvider has been implemented for EReferences and EEnumns, only .");

	}

	/**
	 * Helper method for
	 * {@link #createEditingSupport4ValueProviderBase(ValueProviderBase, ColumnViewer, Class, IAnnotatedSpecification, String)}
	 * that initialized the current input map of the value provider.
	 */
	@SuppressWarnings("unchecked")
	private static <T extends IAnnotatedSpecification, V extends ValueProviderBase<T>> void
			initializeInputChoice(V valueProvider, ColumnViewer viewer,
					Class<? extends IAnnotatedSpecification> clazz, String instanceKey)
					throws Exception {
		// Input choice is not set, or it is restricted to concrete set of values
		// (use SingleEnumAttributeValueProviderBase for Enum types!)
		if(valueProvider.getFixedInputChoiceMap().get(instanceKey) == null ||
				!valueProvider.getFixedInputChoiceMap().get(instanceKey).isEmpty()) {
			valueProvider.getCurrentInputChoiceMap().put(instanceKey,
					valueProvider.getFixedInputChoiceMap().get(instanceKey));

		} else if(!(viewer.getInput() instanceof Set<?>) || ((Set<?>)viewer.getInput()).isEmpty() ||
				!(((Set<?>)viewer.getInput()).iterator().next() instanceof AnnotationEntry)) {
			valueProvider.getCurrentInputChoiceMap().put(instanceKey, null);

		} else {

			AnnotationEntry annotationEntry =
					((Set<AnnotationEntry>)viewer.getInput()).iterator().next();

			EObject root = annotationEntry.getModelElement();
			while(root.eContainer() != null) {
				root = root.eContainer();
			}

			Set<String> choices = new TreeSet<String>();
			for(T specification : (EList<T>)EcoreUtils.getChildrenWithType(root, clazz)) {
				choices.add(valueProvider.getAnnotationValue(specification).toString());
			}

			valueProvider.getCurrentInputChoiceMap().put(instanceKey, choices);
		}
	}

	/**
	 * Creates the {@link EditingSupport} for value providers that are directly based on
	 * {@link ValueProviderBase}.
	 */
	private static <T extends IAnnotatedSpecification, V extends ValueProviderBase<T>>
			EditingSupport createEditingSupport4ValueProviderBase(V valueProvider,
					ColumnViewer viewer, Class<? extends IAnnotatedSpecification> clazz,
					T specification, String instanceKey) throws Exception {

		try {
			initializeInputChoice(valueProvider, viewer, clazz, instanceKey);
		} catch(Exception e) {
			// Use standard text editor as fallback
			valueProvider.setCurrentInputChoiceMap(null);
		}

		// Input is not restricted -> free editing in text cell
		if(valueProvider.getCurrentInputChoiceMap().get(instanceKey) == null) {

			if(valueProvider.getAnnotationValue(specification, instanceKey) instanceof Boolean) {
				return new CheckBoxEditingSupport(viewer, clazz, instanceKey);
			}

			return new TextEditingSupport(viewer, clazz, instanceKey);
		}

		// Multiple instances are not supported (yet) for input to be edited in Comboboxes
		assert (instanceKey == null);

		// Input is restricted to concrete set of values
		// (use SingleEnumAttributeValueProviderBase for Enum types!)
		return new ComboBoxEditingSupport(viewer, clazz, instanceKey, valueProvider
				.getCurrentInputChoiceMap().get(instanceKey), valueProvider
				.getFixedInputChoiceMap().get(instanceKey).isEmpty());
	}

	/**
	 * Creates the editing support.
	 */
	@SuppressWarnings({"rawtypes", "unchecked"})
	public static <T extends IAnnotatedSpecification> EditingSupport createEditingSupport(
			IAnnotationValueProvider<T> valueProvider, ColumnViewer viewer,
			Class<? extends IAnnotatedSpecification> clazz, T specification, String instanceKey)
			throws Exception {

		if(valueProvider instanceof DerivedAnnotationValueProviderBase<?>) {
			return createEditingSupport4DerivedAnnotationValueProviderBase(
					(DerivedAnnotationValueProviderBase)valueProvider, viewer, clazz,
					(IDerivedAnnotation)specification, instanceKey);
		} else if(valueProvider instanceof DynamicInstanceAnnotationValueProviderBase<?>) {
			return createEditingSupport4DynamicInstanceAnnotationValueProviderBase(
					(DynamicInstanceAnnotationValueProviderBase)valueProvider, viewer, clazz,
					specification, instanceKey);
		} else if(valueProvider instanceof EStructuralFeatureValueProviderBase<?>) {
			return createEditingSupport4EStructuralFeatureValueProviderBase(
					(EStructuralFeatureValueProviderBase<T>)valueProvider, viewer, clazz,
					specification, instanceKey);
		}
		if(valueProvider instanceof ValueProviderBase<?>) {
			return createEditingSupport4ValueProviderBase((ValueProviderBase<T>)valueProvider,
					viewer, clazz, specification, instanceKey);
		}

		return null;
	}

	/**
	 * Creates the editing support without referring to the optional {@code instanceKey} (see
	 * {@link #createEditingSupport(IAnnotationValueProvider, ColumnViewer, Class, IAnnotatedSpecification, String)}
	 * .
	 */
	public static <T extends IAnnotatedSpecification> EditingSupport createEditingSupport(
			IAnnotationValueProvider<T> valueProvider, ColumnViewer viewer,
			Class<? extends IAnnotatedSpecification> clazz) throws Exception {

		return createEditingSupport(valueProvider, viewer, clazz, null, null);
	}
}
