/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2014 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.annotation.view.generic.filter;

import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.jface.viewers.ViewerFilter;
import org.fortiss.tooling.base.annotation.AnnotationEntry;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.base.ui.annotation.view.generic.ColumnHandle;
import org.fortiss.tooling.base.ui.annotation.view.generic.GenericAnnotationView;
import org.fortiss.tooling.kernel.model.INamedElement;
import org.fortiss.tooling.kernel.model.IProjectRootElement;

/**
 * Row and column filter for the {@link GenericAnnotationView}.
 * <ul>
 * <li>Row filtering is realized by inheriting {@link ViewerFilter} and the default filtering
 * mechanism provided by {@link TableViewer}.</li>
 * <li>Column filtering is supported by {@link AnnotationFilter#passesColumnFilter(ColumnHandle)}
 * that is evaluated during the update of the {@link GenericAnnotationView}.</li>
 * </ul>
 */
public class AnnotationFilter extends ViewerFilter {
	/** Filter hint text. */
	static final String FILTER_HINT_TEXT = "type filter text";

	/** Hierarchy filter option: all model elements from any level */
	static final String HIERARCHY_LEVELS_ALL = "Show all levels";

	/** Hierarchy filter option: all model elements from current level */
	static final String HIERARCHY_LEVELS_CURRENT = "Show current level";

	/** Hierarchy filter option: currently selected model element and its offspring */
	static final String HIERARCHY_LEVELS_SELECTED_SUBMODEL = "Show selected sub-model";

	/** Current model element / annotation name filter pattern */
	private String filterPattern;

	/** Flag whether columns / annotation names should be filtered. */
	private boolean filterColumnName;

	/** Flag whether rows / model element names should be filtered */
	private boolean filterRowName;

	/** Currently selected model element (may be {@code null}). */
	private IModelElement currentySelectedModelElement;

	/**
	 * Flag whether the displayed rows should be restricted to the currently selected model
	 * element type.
	 */
	boolean restrictToSelectedModelElementType;

	/** Row filter option based on hierarchy level of model element. */
	private String hierarchyLevelFilter;

	/** Column filter option based on annotation type */
	private Class<? extends IAnnotatedSpecification> annotationTypeFilter;

	/** Flag if a full update of the annotation view has been requested. */
	boolean annotationViewFullUpdateRequested;

	/** {@inheritDoc} */
	@Override
	public boolean select(Viewer viewer, Object parentElement, Object element) {
		if(!(element instanceof AnnotationEntry)) {
			return true;
		}

		AnnotationEntry annotationEntry = ((AnnotationEntry)element);
		if(!(annotationEntry.getModelElement() instanceof INamedElement)) {
			return true;
		}

		return passesRowFilter(annotationEntry);
	}

	/**
	 * Returns {@code true} if a given {@code input} passes an case-insensitive filter specified
	 * by {@code filterString}.
	 */
	private boolean passesNameFilter(String input) {
		// Null-filter accepts every input
		if(filterPattern == null || filterPattern.equals(FILTER_HINT_TEXT)) {
			return true;
		}

		// Null-input cannot be accepted by a non-null filter
		if(input == null) {
			return false;
		}
		return input.toLowerCase().contains(filterPattern.toLowerCase());
	}

	/**
	 * Returns {@code true} if given annotation entry matches the currently selected model
	 * element (or if this particular filter option is turned off).
	 */
	private boolean passesSelectedElementTypeFilter(AnnotationEntry annotationEntry) {
		if(currentySelectedModelElement == null ||
				currentySelectedModelElement instanceof IProjectRootElement ||
				hierarchyLevelFilter == null) {
			return true;
		}

		int currentlySelectedModelElementLevel =
				GenericAnnotationView.getModelElementLevel(currentySelectedModelElement);
		int modelElementLevel =
				GenericAnnotationView.getModelElementLevel(annotationEntry.getModelElement());

		if(hierarchyLevelFilter.equals(HIERARCHY_LEVELS_CURRENT)) {
			if((GenericAnnotationView.getModelElementLevel(annotationEntry.getModelElement()) != currentlySelectedModelElementLevel) ||
					(currentySelectedModelElement.eContainer() != GenericAnnotationView
							.getModelElementAncestor(annotationEntry.getModelElement(),
									modelElementLevel - (currentlySelectedModelElementLevel - 1)))) {
				return false;
			}
		}
		if(hierarchyLevelFilter.equals(HIERARCHY_LEVELS_SELECTED_SUBMODEL)) {
			if(modelElementLevel < currentlySelectedModelElementLevel) {
				return false;
			}

			if((modelElementLevel > currentlySelectedModelElementLevel) &&
					((currentySelectedModelElement != GenericAnnotationView
							.getModelElementAncestor(annotationEntry.getModelElement(),
									modelElementLevel - currentlySelectedModelElementLevel))) ||
					(currentySelectedModelElement != GenericAnnotationView.getModelElementAncestor(
							annotationEntry.getModelElement(), modelElementLevel -
									(currentlySelectedModelElementLevel - 0)))) {
				return false;
			}

		}

		return !restrictToSelectedModelElementType ||
				annotationEntry.getModelElement().getClass()
						.equals(currentySelectedModelElement.getClass());
	}

	/** Returns true if the given specification passes the current annotation type filter */
	private boolean passesAnnotationTypeFilter(IAnnotatedSpecification spec) {
		return annotationTypeFilter == null || spec.getClass().equals(annotationTypeFilter);
	}

	/**
	 * Returns {@code true} if the column represented by the given column is visible in the
	 * view.
	 */
	public boolean passesColumnFilter(ColumnHandle columnHandle) {
		return (!filterColumnName || passesNameFilter(columnHandle.getColumnName())) &&
				passesSelectedElementTypeFilter(columnHandle.getEntry()) &&
				passesAnnotationTypeFilter(columnHandle.getAnnotatedSpecification());
	}

	/**
	 * Returns {@code true} if the rows represented by {@code annotationEntry} are visible
	 * in the view.
	 */
	public boolean passesRowFilter(AnnotationEntry annotationEntry) {
		if(!(annotationEntry.getModelElement() instanceof INamedElement)) {
			return true;
		}

		return (!filterRowName || passesNameFilter(((INamedElement)annotationEntry
				.getModelElement()).getName())) && passesSelectedElementTypeFilter(annotationEntry);
	}

	/**
	 * Query if a full update of the annotation view as been requested (and reset the request), or
	 * if the current filter settings require a full update in any case.
	 */
	public boolean annotationViewQueryUpdateRequired() {
		// In case an update has been explicitly requested, reset the request here.
		boolean annotationViewFullUpdateRequested = this.annotationViewFullUpdateRequested;
		this.annotationViewFullUpdateRequested = false;

		// If the filter is based on the current selection (i.e., currently selected model element
		// type, current level, or current sub-model, ), a full update is required on very
		// selection change since the newly selected element might contribute a different set of
		// annotations.
		return restrictToSelectedModelElementType ||
				!hierarchyLevelFilter.equals(HIERARCHY_LEVELS_ALL) ||
				annotationViewFullUpdateRequested;
	}

	/** Request a full update of the annotation view (due to changed filter settings). */
	public void annotationViewRequestFullUpdate() {
		this.annotationViewFullUpdateRequested = true;
	}

	/** Sets the model element / annotation name filter pattern. */
	public void setNameFilterPattern(String filterPattern) {
		this.filterPattern = filterPattern;
	}

	/** Sets the flag whether rows should be filtered by name. */
	public void setFilterRowName(boolean filterRowName) {
		this.filterRowName = filterRowName;
	}

	/** Sets the flag whether columns should be filtered by name. */
	public void setFilterColumnName(boolean filterColumnName) {
		this.filterColumnName = filterColumnName;
	}

	/** Sets the flag whether the view should be filtered by the selected model element type. */
	public void setRestrictToSelectedModelElementType(boolean restrictToSelectedModelElementType) {
		this.restrictToSelectedModelElementType = restrictToSelectedModelElementType;
	}

	/** Sets the model element hierarchy level filter option. */
	public void setHierarchyLevelFilter(String hierarchyLevelFilter) {
		this.hierarchyLevelFilter = hierarchyLevelFilter;
	}

	/** Sets the annotation type filter option. */
	public void setAnnotationTypeFilter(
			Class<? extends IAnnotatedSpecification> annotationTypeFilter) {
		this.annotationTypeFilter = annotationTypeFilter;
	}

	/** Sets the currently selected model element (may be {@code null}). */
	public void setCurrentySelectedModelElement(IModelElement currentySelectedModelElement) {
		this.currentySelectedModelElement = currentySelectedModelElement;
	}
}
