/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2014 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.annotation.valueprovider;

import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IDerivedAnnotation;
import org.fortiss.tooling.base.model.element.IModelElement;

/**
 * Description of a {@link EStructuralFeature} managed by a
 * {@link EStructuralFeatureValueProviderBase}.
 * 
 * @author barner
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating RED Hash:
 */
public class EStructuralFeatureDescriptor {
	/**
	 * Scope in which {@link EReference}s managed by {@link EStructuralFeatureDescriptor} are
	 * resolved.
	 */
	public enum EReferenceScope {
		/** Current project (all models) */
		PROJECT,

		/** Current model (e.g., ComponentArchitecture) */
		MODEL,

		/**
		 * Sub-model below the {@link IModelElement} to which the annotation is attached that
		 * contains the {@link EStructuralFeature} managed by this
		 * {@link EStructuralFeatureDescriptor} (including the element itself)
		 */
		SUB_MODEL,

		/**
		 * Sub-model below the {@link IModelElement} to which the annotation is attached that
		 * contains the {@link EStructuralFeature} managed by this
		 * {@link EStructuralFeatureDescriptor} (NOT including the element itself)
		 */
		STRICT_SUB_MODEL
	}

	/** {@link EStructuralFeature} managed by this {@link EStructuralFeatureDescriptor}. */
	private EStructuralFeature eStructuralFeature;

	/**
	 * Scope in which {@link EReference}s managed by {@link EStructuralFeatureDescriptor} are
	 * resolved ({@code null} in case the {@link EStructuralFeature} managed by this
	 * {@link EStructuralFeatureDescriptor} is not an {@link EReference}).
	 */
	private EReferenceScope eReferenceScope;

	/** Constructs a {@link EStructuralFeatureDescriptor} for an {@link EAttribute}. */
	public EStructuralFeatureDescriptor(EAttribute eAttribute) {
		this.eStructuralFeature = eAttribute;
		this.eReferenceScope = null;
	}

	/** Constructs a {@link EStructuralFeatureDescriptor} for an {@link EReference}. */
	public EStructuralFeatureDescriptor(EReference eReference, EReferenceScope eReferenceScope) {
		this.eStructuralFeature = eReference;
		this.eReferenceScope = eReferenceScope;
	}

	/**
	 * Returns the {@link EStructuralFeature} described by this {@link EStructuralFeatureDescriptor}
	 * .
	 * 
	 * @param specification
	 */
	public EStructuralFeature getEStructuralFeature(IAnnotatedSpecification specification) {
		// IDerivedAnnotation supports dynamically switching the return type of getValue() (the
		// calculated value)
		if(specification instanceof IDerivedAnnotation) {
			EStructuralFeature derivedFeature =
					((IDerivedAnnotation<?>)specification).getDerivedFeature();
			if(derivedFeature != null) {
				return derivedFeature;
			}
		}

		return eStructuralFeature;
	}

	/**
	 * Returns the scope in which {@link EReference}s managed by
	 * {@link EStructuralFeatureDescriptor} are resolved ({@code null} in case the
	 * {@link EStructuralFeature} managed by this {@link EStructuralFeatureDescriptor} is not an
	 * {@link EReference}).
	 */
	public EReferenceScope getEReferenceScope() {
		return eReferenceScope;
	}

	/**
	 * Predicate if an object is available to be selected for an {@link EReference} structural
	 * feature of an annotation. If this feature is required, concrete
	 * {@link IAnnotationValueProvider}s can provide a derived {@link EStructuralFeatureDescriptor}
	 * that overrides this method.
	 * 
	 * @param obj
	 *            Object to be checked for availability.
	 * @param modelElement
	 *            Model element to which {@code specification} is attached to
	 * @param specification
	 *            Specification for which the set of available objects is computed.
	 * 
	 */
	public boolean isAvailableObject(EObject obj, IAnnotatedSpecification specification,
			EObject modelElement) {
		return true;
	}

	/**
	 * Returns the type of the structural feature managed by this
	 * {@link EStructuralFeatureDescriptor}.
	 */
	public EClassifier getEType(IAnnotatedSpecification specification) {
		EClassifier eType = null;
		try {
			eType = getEStructuralFeature(specification).getEType();
		} catch(Exception e) {
			// Ignore
		}
		return eType;
	}

}
