/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2011 ForTISS GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.ui.internal;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.jface.action.IContributionItem;
import org.eclipse.jface.action.IMenuListener;
import org.eclipse.jface.action.IMenuManager;
import org.eclipse.jface.action.MenuManager;
import org.eclipse.jface.action.Separator;
import org.eclipse.ui.IWorkbenchActionConstants;
import org.fortiss.tooling.kernel.ToolingKernelActivator;
import org.fortiss.tooling.kernel.ui.extension.IContextMenuContributor;
import org.fortiss.tooling.kernel.ui.extension.IContextMenuMultiSelectionContributor;
import org.fortiss.tooling.kernel.ui.extension.data.ContextMenuContextProvider;
import org.fortiss.tooling.kernel.ui.service.IActionService;
import org.fortiss.tooling.kernel.ui.service.IContextMenuService;
import org.fortiss.tooling.kernel.utils.ExtensionPointUtils;
import org.fortiss.tooling.kernel.utils.LoggingUtils;
import org.osgi.framework.Bundle;

/**
 * This class implements the {@link IContextMenuService} interface.
 * 
 * @author hoelzl
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating GREEN Hash: C768848649422B67FE033EFFD57F7A04
 */
public class ContextMenuService implements IContextMenuService {

	/** The model element handler extension point ID. */
	private static final String EXTENSION_POINT_NAME =
			"org.fortiss.tooling.kernel.ui.contextMenuContribution";

	/** The model element handler configuration element name. */
	private static final String CONFIGURATION_ELEMENT_NAME = "contextMenuContribution";

	/** The model element handler class attribute name. */
	private static final String HANDLER_CLASS_ATTRIBUTE_NAME = "contributor";

	/** Stores the registered prototype providers. */
	private final List<IContextMenuContributor> providerList =
			new ArrayList<IContextMenuContributor>();

	/** Constructor. */
	public ContextMenuService() {
		setupProviders();
	}

	/** {@inheritDoc} */
	@Override
	public MenuManager createDefaultContextMenu(ContextMenuContextProvider contextProvider) {
		MenuManager menuManager = new MenuManager("#PopupMenu");
		menuManager.setRemoveAllWhenShown(true);
		menuManager.addMenuListener(new MenuCreator(contextProvider));

		return menuManager;
	}

	/** Initializes the prototype list from plugin extensions. */
	private void setupProviders() {
		for(IConfigurationElement ce : ExtensionPointUtils.getConfigurationElements(
				EXTENSION_POINT_NAME, CONFIGURATION_ELEMENT_NAME)) {
			Bundle bundle = ExtensionPointUtils.getBundle(ce);
			try {
				Class<?> handlerClass =
						ExtensionPointUtils.loadClass(
								ce.getAttribute(HANDLER_CLASS_ATTRIBUTE_NAME), bundle);

				IContextMenuContributor provider =
						(IContextMenuContributor)handlerClass.getConstructor().newInstance();
				providerList.add(provider);

			} catch(Exception ex) {
				LoggingUtils.error(ToolingKernelActivator.getDefault(), ex.getMessage(), ex);
			}
		}
	}

	/** Wrapper class for menu creation. */
	private class MenuCreator implements IMenuListener {

		/** Stores the context object. */
		private final ContextMenuContextProvider contextProvider;

		/** Constructor. */
		public MenuCreator(ContextMenuContextProvider contextProvider) {
			this.contextProvider = contextProvider;
		}

		/** {@inheritDoc} */
		@Override
		public void menuAboutToShow(IMenuManager menu) {
			if(!menu.isEmpty()) {
				return;
			}
			addVisibleSectionSeparator(menu, TOP_MOST_MENU_SECTION_ID);
			addVisibleSectionSeparator(menu, BEFORE_GLOBAL_MENU_SECTION_ID);
			addVisibleSectionSeparator(menu, ActionService.GLOBAL_DEFAULT_MENU_SECTION_ID);
			addVisibleSectionSeparator(menu, AFTER_GLOBAL_MENU_SECTION_ID);
			addVisibleSectionSeparator(menu, REPOSITORY_MENU_SECTION_ID);
			addVisibleSectionSeparator(menu, IWorkbenchActionConstants.MB_ADDITIONS);
			addVisibleSectionSeparator(menu, BOTTOM_MOST_MENU_SECTION_ID);

			addContributions(menu, contextProvider);

			IActionService.INSTANCE.addGlobalDefaultActionSectionToMenu(menu);
		}
	}

	/** Adds registered contributions to the given menu. */
	private void addContributions(IMenuManager menu, ContextMenuContextProvider contextProvider) {

		List<EObject> selection = contextProvider.getSelectedModelElementList();
		if(selection == null) {
			// do nothing
		} else if(selection.size() == 1) {
			EObject selectionElem = contextProvider.getSelectedModelElement();
			for(IContextMenuContributor contributor : providerList) {
				String menuSectionID = contributor.getMenuSectionID();
				if(menuSectionID == null) {
					menuSectionID = IWorkbenchActionConstants.MB_ADDITIONS;
				}
				List<IContributionItem> l =
						contributor.getContributedItems(selectionElem, contextProvider);
				if(l == null) {
					continue;
				}
				for(IContributionItem item : l) {
					menu.appendToGroup(menuSectionID, item);
				}
			}
		} else if(selection.size() > 1) {
			for(IContextMenuContributor contributor : providerList) {
				if(contributor instanceof IContextMenuMultiSelectionContributor) {
					IContextMenuMultiSelectionContributor multicontributor =
							(IContextMenuMultiSelectionContributor)contributor;
					String menuSectionID = contributor.getMenuSectionID();
					if(menuSectionID == null) {
						menuSectionID = IWorkbenchActionConstants.MB_ADDITIONS;
					}
					List<IContributionItem> l =
							multicontributor.getContributedItems(selection, contextProvider);
					if(l == null) {
						continue;
					}
					for(IContributionItem item : l) {
						menu.appendToGroup(menuSectionID, item);
					}
				}
			}
		}
	}

	/** Adds a visible section separator. */
	private void addVisibleSectionSeparator(IMenuManager menu, String sectionName) {
		Separator separator = new Separator(sectionName);
		separator.setVisible(true);
		menu.add(separator);
	}
}
