/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2011 ForTISS GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.ui.internal;

import static org.fortiss.tooling.kernel.utils.KernelModelElementUtils.getParentElement;
import static org.fortiss.tooling.kernel.utils.LoggingUtils.error;

import java.util.EventObject;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.conqat.lib.commons.collections.CollectionUtils;
import org.eclipse.core.runtime.Assert;
import org.eclipse.emf.common.command.CommandStackListener;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IPartListener;
import org.eclipse.ui.IPartService;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.PlatformUI;
import org.fortiss.tooling.kernel.ToolingKernelActivator;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.model.ILibraryElement;
import org.fortiss.tooling.kernel.model.ILibraryElementReference;
import org.fortiss.tooling.kernel.service.ICommandStackService;
import org.fortiss.tooling.kernel.service.ILibraryService;
import org.fortiss.tooling.kernel.service.IPersistencyService;
import org.fortiss.tooling.kernel.service.base.EObjectAwareServiceBase;
import org.fortiss.tooling.kernel.ui.extension.IModelEditor;
import org.fortiss.tooling.kernel.ui.extension.IModelEditorBinding;
import org.fortiss.tooling.kernel.ui.extension.IModelElementHandler;
import org.fortiss.tooling.kernel.ui.internal.editor.ExtendableMultiPageEditor;
import org.fortiss.tooling.kernel.ui.internal.editor.ModelElementEditorInput;
import org.fortiss.tooling.kernel.ui.service.IModelEditorBindingService;
import org.fortiss.tooling.kernel.ui.service.IModelElementHandlerService;
import org.fortiss.tooling.kernel.ui.service.INavigatorService;
import org.fortiss.tooling.kernel.utils.KernelModelElementUtils;

/**
 * This class implements the {@link IModelEditorBindingService} interface.
 * 
 * @author hoelzl
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating GREEN Hash: 8D45C2B6E2FC7158EF002A9ECE848DD2
 */
public class ModelEditorBindingService extends
		EObjectAwareServiceBase<IModelEditorBinding<EObject>> implements
		IModelEditorBindingService, IPartListener, CommandStackListener {

	/** The compositor extension point ID. */
	private static final String EXTENSION_POINT_NAME =
			"org.fortiss.tooling.kernel.ui.modelEditorBinding";

	/** The compositor configuration element name. */
	private static final String CONFIGURATION_ELEMENT_NAME = "modelEditorBinding";

	/** The handler class attribute name. */
	private static final String HANDLER_CLASS_ATTRIBUTE_NAME = "binding";

	/** Stores the currently opened editors and their model elements. */
	private final Map<IEditorPart, EObject> currentEditors = new HashMap<IEditorPart, EObject>();

	/** Stores the command stack registrations. */
	private final Map<EObject, ITopLevelElement> currentCommandStacks =
			new HashMap<EObject, ITopLevelElement>();

	/** Constructor. */
	public ModelEditorBindingService() {
		super();
	}

	/** {@inheritDoc} */
	@Override
	public void openInEditor(EObject element) {
		openInEditor(element, 2);
	}

	/**
	 * Opens the editor or proceeds with parent element.
	 * Particular treatment for library references which open directly the library element.
	 */
	private void openInEditor(EObject element, int depth) {
		Assert.isTrue(depth >= 0);
		ILibraryElementReference libRefParent =
				getParentElement(element, ILibraryElementReference.class);
		if(libRefParent != null) {
			String uri = libRefParent.getURI();
			ILibraryElement libElement = ILibraryService.INSTANCE.findLibraryElementByURI(uri);
			IModelElementHandler<EObject> handler =
					IModelElementHandlerService.INSTANCE.getModelElementHandler(libElement);
			if(handler != null) {
				EObject displayedElement = handler.handleOpenModelElementRequest(libElement);
				INavigatorService.INSTANCE.setLibraryViewActive();
				openInEditor(displayedElement);
			}
			return;
		}
		if(getBindings(element).isEmpty()) {
			if(depth > 0 && element.eContainer() != null) {
				openInEditor(element.eContainer(), depth - 1);
			}
			return;
		}
		try {
			IEditorPart part =
					PlatformUI
							.getWorkbench()
							.getActiveWorkbenchWindow()
							.getActivePage()
							.openEditor(new ModelElementEditorInput(element),
									ExtendableMultiPageEditor.ID);
			currentEditors.put(part, element);
			// ensure registration with part service
			IPartService service = (IPartService)part.getSite().getService(IPartService.class);
			service.addPartListener(new EditorSitePartListener());
			// ensure registration with command stack service
			ITopLevelElement top = IPersistencyService.INSTANCE.getTopLevelElementFor(element);
			ICommandStackService.INSTANCE.addCommandStackListener(top, this);
			currentCommandStacks.put(element, top);
		} catch(final PartInitException e) {
			error(ToolingKernelActivator.getDefault(), "Could not open editor with ID " +
					ExtendableMultiPageEditor.ID, e);
		}
	}

	/** {@inheritDoc} */
	@Override
	public List<IModelEditorBinding<EObject>> getBindings(EObject element) {
		List<IModelEditorBinding<EObject>> bindings = getRegisteredHandlers(element.getClass());
		if(bindings == null) {
			bindings = CollectionUtils.emptyList();
		}
		return bindings;
	}

	/** {@inheritDoc} */
	@Override
	protected String getExtensionPointName() {
		return EXTENSION_POINT_NAME;
	}

	/** {@inheritDoc} */
	@Override
	protected String getConfigurationElementName() {
		return CONFIGURATION_ELEMENT_NAME;
	}

	/** {@inheritDoc} */
	@Override
	protected String getHandlerClassAttribute() {
		return HANDLER_CLASS_ATTRIBUTE_NAME;
	}

	/** {@inheritDoc} */
	@Override
	public void closeEditors(EObject parentElement) {
		IEditorPart[] parts = new IEditorPart[0];
		for(IEditorPart editor : currentEditors.keySet().toArray(parts)) {
			EObject editedElement = currentEditors.get(editor);
			if(KernelModelElementUtils.isChildElementOf(editedElement, parentElement)) {
				closeEditor(editor);
			}
		}

	}

	/** {@inheritDoc} */
	@Override
	public void closeEditor(EObject element) {
		IEditorPart[] parts = new IEditorPart[0];
		for(IEditorPart editor : currentEditors.keySet().toArray(parts)) {
			EObject editedElement = currentEditors.get(editor);
			if(editedElement.equals(element)) {
				closeEditor(editor);
			}
		}

	}

	/** Closes the given editor. */
	private void closeEditor(final IEditorPart editor) {
		editor.getSite().getShell().getDisplay().asyncExec(new Runnable() {
			@Override
			public void run() {
				editor.getSite().getPage().closeEditor(editor, false);
			}
		});
	}

	/** {@inheritDoc} */
	@Override
	public void partActivated(IWorkbenchPart part) {
		// ignore
	}

	/** {@inheritDoc} */
	@Override
	public void partBroughtToTop(IWorkbenchPart part) {
		// ignore
	}

	/** {@inheritDoc} */
	@Override
	public void partClosed(IWorkbenchPart part) {
		if(currentEditors.containsKey(part)) {
			EObject eo = currentEditors.get(part);
			ITopLevelElement top = currentCommandStacks.get(eo);
			currentCommandStacks.remove(eo);
			currentEditors.remove(part);

			// clean up command stack registration
			for(IEditorPart other : currentEditors.keySet()) {
				if(top == currentCommandStacks.get(currentEditors.get(other))) {
					return;
				}
			}
			// no other editor references the command stack of top anymore
			ICommandStackService.INSTANCE.removeCommandStackListener(top, this);
		}
	}

	/** {@inheritDoc} */
	@Override
	public void partDeactivated(IWorkbenchPart part) {
		// ignore
	}

	/** {@inheritDoc} */
	@Override
	public void partOpened(IWorkbenchPart part) {
		// can be ignored
	}

	/** {@inheritDoc} */
	@Override
	public void commandStackChanged(EventObject event) {
		IEditorPart[] parts = new IEditorPart[0];
		for(IEditorPart part : currentEditors.keySet().toArray(parts)) {
			ITopLevelElement top =
					IPersistencyService.INSTANCE.getTopLevelElementFor(currentEditors.get(part));
			if(top == null) {
				closeEditor(part);
			}
		}
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public IModelEditor<EObject> getActiveEditor() {
		IEditorPart activeEditor =
				PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage()
						.getActiveEditor();
		if(activeEditor instanceof ExtendableMultiPageEditor) {
			ExtendableMultiPageEditor extEditor = (ExtendableMultiPageEditor)activeEditor;
			IModelEditor<EObject> inner = extEditor.getActiveModelEditor();
			if(inner != null) {
				return inner;
			}
			return extEditor;
		}
		if(activeEditor instanceof IModelEditor) {
			return (IModelEditor<EObject>)activeEditor;
		}
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public boolean isOpen(EObject element) {
		return currentEditors.containsValue(element);
	}

	/**
	 * Instance class used for listening to editor part life-cycle. This class is needed, because
	 * Eclipse removes this instance once the editor site is disposed. Therefore, we cannot use the
	 * service singleton directly, since multiple editors would add it, but the first closed one
	 * removes it.
	 */
	private class EditorSitePartListener implements IPartListener {

		/** {@inheritDoc} */
		@Override
		public void partActivated(IWorkbenchPart part) {
			ModelEditorBindingService.this.partActivated(part);
		}

		/** {@inheritDoc} */
		@Override
		public void partBroughtToTop(IWorkbenchPart part) {
			ModelEditorBindingService.this.partBroughtToTop(part);
		}

		/** {@inheritDoc} */
		@Override
		public void partClosed(IWorkbenchPart part) {
			ModelEditorBindingService.this.partClosed(part);
		}

		/** {@inheritDoc} */
		@Override
		public void partDeactivated(IWorkbenchPart part) {
			ModelEditorBindingService.this.partDeactivated(part);
		}

		/** {@inheritDoc} */
		@Override
		public void partOpened(IWorkbenchPart part) {
			ModelEditorBindingService.this.partOpened(part);
		}
	}
}
